package ru.dbotthepony.mc.otm.client.render

import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.client.renderer.texture.SpriteContents
import net.minecraft.client.renderer.texture.atlas.SpriteResourceLoader
import net.minecraft.client.renderer.texture.atlas.SpriteSource
import net.minecraft.client.renderer.texture.atlas.SpriteSource.TEXTURE_ID_CONVERTER
import net.minecraft.client.renderer.texture.atlas.SpriteSourceType
import net.minecraft.client.renderer.texture.atlas.sources.LazyLoadedImage
import net.minecraft.client.resources.metadata.animation.FrameSize
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.packs.resources.ResourceManager
import net.minecraft.server.packs.resources.ResourceMetadata
import net.minecraft.world.item.DyeColor
import net.neoforged.neoforge.client.event.RegisterSpriteSourceTypesEvent
import net.neoforged.neoforge.client.model.generators.ModelProvider
import net.neoforged.neoforge.common.data.ExistingFileHelper
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.mc.otm.OverdriveThatMatters.loc

class ColorizedSpriteSource(val base: ResourceLocation, val overlay: ResourceLocation, val outputPathBase: ResourceLocation, val outputPathSuffix: String = "/", val tintBase: Boolean = false, val tintOverlay: Boolean = true) : SpriteSource {
	override fun run(
		manager: ResourceManager,
		output: SpriteSource.Output
	) {
		val baseRes = manager.getResource(TEXTURE_ID_CONVERTER.idToFile(base)).get()
		val maskRes = manager.getResource(TEXTURE_ID_CONVERTER.idToFile(overlay)).get()

		for (baseTint in DyeColor.entries) {
			for (overlayTint in DyeColor.entries) {
				if (tintBase && baseTint == overlayTint) continue

				var outputLocation = outputPathBase
				if (tintBase) outputLocation = outputLocation.withSuffix(outputPathSuffix + baseTint.serializedName)
				if (tintOverlay) outputLocation = outputLocation.withSuffix(outputPathSuffix + overlayTint.serializedName)

				val baseImage = LazyLoadedImage(base, baseRes, 1)
				val maskImage = LazyLoadedImage(overlay, maskRes, 1)

				output.add(outputLocation, Supplier(baseImage, if (tintBase) baseTint else null, maskImage, if (tintOverlay) overlayTint else null, outputLocation))
			}

			if (!tintBase) return
		}
	}

	override fun type(): SpriteSourceType {
		return SPRITE_SOURCE!!
	}

	fun trackGeneratedTextures(efh: ExistingFileHelper) : ColorizedSpriteSource {
		for (baseTint in DyeColor.entries) {
			for (overlayTint in DyeColor.entries) {
				if (tintBase && baseTint == overlayTint) continue

				var outputLocation = outputPathBase
				if (tintBase) outputLocation = outputLocation.withSuffix(outputPathSuffix + baseTint.serializedName)
				if (tintOverlay) outputLocation = outputLocation.withSuffix(outputPathSuffix + overlayTint.serializedName)

				efh.trackGenerated(outputLocation, ModelProvider.TEXTURE)
			}

			if (!tintBase) return this
		}

		return this
	}

	class Supplier(val baseImage: LazyLoadedImage, val baseTint: DyeColor?, val maskImage: LazyLoadedImage, val maskTint: DyeColor?, val outputLocation: ResourceLocation) : SpriteSource.SpriteSupplier {
		override fun apply(t: SpriteResourceLoader?): SpriteContents? {
			val base = baseImage.get()
			val mask = maskImage.get()

			if (base.width != mask.width || base.height != mask.height) {
				throw IllegalArgumentException("Base image doesn't match mask in texture size")
			}

			if (baseTint != null) {
				val baseDiffCol = RGBAColor.argb(baseTint.textureDiffuseColor)
				base.applyToAllPixels {
					val a = (it and -0x1000000 ushr 24) / 255f
					val b = (it and 0xFF0000 ushr 16) / 255f
					val g = (it and 0xFF00 ushr 8) / 255f
					val r = (it and 0xFF) / 255f

					RGBAColor(r * baseDiffCol.red, g * baseDiffCol.green, b * baseDiffCol.blue, a).toRGBA()
				}
			}

			if (maskTint != null) {
				val maskDiffCol = RGBAColor.argb(maskTint.textureDiffuseColor)
				mask.applyToAllPixels {
					val a = (it and -0x1000000 ushr 24) / 255f
					val b = (it and 0xFF0000 ushr 16) / 255f
					val g = (it and 0xFF00 ushr 8) / 255f
					val r = (it and 0xFF) / 255f

					RGBAColor(r * maskDiffCol.red, g * maskDiffCol.green, b * maskDiffCol.blue, a).toRGBA()
				}
			}

			for (i in 0 until base.width) {
				for (j in 0 until base.height) {
					base.blendPixel(i, j, mask.getPixelRGBA(i, j))
				}
			}

			maskImage.release()

			return SpriteContents(outputLocation, FrameSize(base.width, base.height), base, ResourceMetadata.EMPTY)
		}

	}

	companion object {
		val LOCATION = loc("colorized")

		private val CODEC: MapCodec<ColorizedSpriteSource> = RecordCodecBuilder.mapCodec {
			it.group(
				ResourceLocation.CODEC.fieldOf("base").forGetter { it.base },
				ResourceLocation.CODEC.fieldOf("overlay").forGetter { it.overlay },
				ResourceLocation.CODEC.fieldOf("output_path_base").forGetter { it.outputPathBase },
				Codec.STRING.optionalFieldOf("output_path_suffix", "/").forGetter { it.outputPathSuffix },
				Codec.BOOL.optionalFieldOf("base_tint", false).forGetter { it.tintBase },
				Codec.BOOL.optionalFieldOf("overlay_tint", true).forGetter{ it.tintOverlay }

			).apply(it, ::ColorizedSpriteSource)
		}

		var SPRITE_SOURCE: SpriteSourceType? = null

		fun register(event: RegisterSpriteSourceTypesEvent) {
			SPRITE_SOURCE = event.register(LOCATION, CODEC)
		}
	}
}
